#include "contrax.h"
#include "matrix2.h"
#include "exact.h"
#include "myops.h"

FILE *infile, *outfile, *treefile;
long numsp, numsp1, numsp2, chars, numtrees, j;
phenotype *x, *cntrast;
naym *nayms;
boolean printdata, progress, reg, mulsets, ibmpc, vt52, ansi,
       bifurcating;
Char ch;

/* Local variables for maketree, propagated globally for c version: */
tree curtree;
long contno;

MAT *full_mat=MNULL;
Real q_sum=0.;
int nrow;

void print_tree(node *p)
    {
    int ic,i;
    if(p==NULL) return;
    if(!p->tip)
	{
	print_tree(p->next->back);
	print_tree(p->next->next->back);
	}
/*  printf("node %ld v %lf view ",p->number, p->v); */
    printf(" %ld       %lf       ",p->number, p->v);

    for(ic=0; ic<chars; ic++)
	printf("%lf ",p->view[ic]);
/*    printf("  %s \n",p->nayme); */
    for (i=0; i<namelength; i++)
	putchar(p->nayme[i]);
    printf("\n");
    }

void fix_zero_v(node *p)
    {
    int ic;
    if(p==NULL) return;
    if(!p->tip)
	{
    	fix_zero_v(p->next->back);
    	fix_zero_v(p->next->next->back);
	}
    if (p->v==0) p->v=0.000000000001;
    }

void get_stats()
    {
    VEC *qhat,*red_qpp,*SE;
    double Q;
    int itry;
    long int df;
    printf("\nTree as read:\n");
    printf("(use 'printree' or Phylip's 'retree' program to view node numbering) \n");
    printf("Node   Branch length  Trait values   Name\n");
    print_tree(curtree.start);
    nrow=(int) (numsp2-numsp1+1);
    
    fix_zero_v(curtree.start);

    qhat = get_qhat(&curtree,VNULL);
    printf("\nAfter ML estimates inserted:\n");
    printf("Node   Branch length  Trait values   Name\n");
    print_tree(curtree.start);
    red_qpp=v_get(nrow);
    SE=v_get(nrow);
    Q=v_sum(qhat);
    df=chars*(numsp2-numsp1+1);
    printf("\nEstimate of beta= %lf  df= %ld \n", (Q/df), df);

    for(itry=0; itry < nrow; itry++){
	red_qpp->ve[itry]= get_t_dist((int)(numsp1+itry));
	SE->ve[itry]=sqrt(2*Q/(df*red_qpp->ve[itry]));
	}
/*    printf("\nred_qpp: "); v_output(red_qpp); */

    printf("\nNode    SE of ancestor state\n");
    for(itry=0; itry < nrow; itry++)
	printf(" %ld          %lf\n",itry+numsp1,SE->ve[itry]);
    printf("\n");

    }

Real get_t_dist(int node_number)
    {
    int ir;
    Real red_qpp, qpp;
    MAT *less_row,*reduced;
    VEC *grab,*dgdp;
    if(numsp1 > node_number || node_number > numsp2)
	{
	fprintf(stderr,
	"get_t_dist passed bad node number: %d\n",node_number);
	exit(1);
	}
    ir=(int) (node_number-numsp1);
    qpp=full_mat->me[ir][ir]; /* save diagonal */
    /*printf("\nmat before deleting row: "); m_output(full_mat);*/
    less_row=m_delete_row(full_mat,ir,MNULL);
    /*printf("\nmat after deleting row %d: ",ir); m_output(less_row);*/
    grab= get_col(less_row,ir,VNULL);
    /*printf("\ngrab col: "); v_output(grab);*/
    reduced=m_delete_col(less_row,ir,MNULL);
    /*printf("\nreduced: "); m_output(reduced);*/
    M_FREE(less_row);
    reduced=CHfactor(reduced);
    dgdp=CHsolve(reduced,grab,VNULL);
    red_qpp = qpp - in_prod(grab,dgdp);
    /*printf("\nqpp %lf red_qpp %lf\n",qpp,red_qpp);*/
    return red_qpp;
    }


VEC *get_qhat(tree *tp,VEC *qhat) /* evaluate the ML residual sum of squares */
    {
    int ic;
    get_full_matrix(tp); /* get ML estimates and insert in nodes */
    qhat=v_resize(qhat,chars);
    for(ic=0; ic<chars; ic++)
	{
	q_sum=0;
    	add_q(tp->start,ic); /* recursive evaluation of the sum of squares */
	qhat->ve[ic]=q_sum;
	}
    /*printf("sum of squares vector: "); v_output(qhat);*/
    return qhat;
    }

void add_q(node *p,int ic)
    {
    if(p->tip) return;
    add_q(p->next->back,ic);
    add_q(p->next->next->back,ic);

    one_q(p,p->next->back,ic);
    one_q(p,p->next->next->back,ic);
    }

void one_q(node *p, node *q,int ic)
    {
    double temp;
	temp =p->view[ic] - q->view[ic];
    q_sum += temp*temp / q->v;
    }


void get_full_matrix(tree *tp)
    {

/* initialize matrices, etc */

    MAT *full_rhs,*ch_fact;
    VEC *mle,*rhs;
    Real **fm_me;
    Real **fv_me;
    int ip,ic;
    node *p,**np;
    full_rhs=MNULL;
    mle=VNULL;
    full_mat=m_resize(full_mat,nrow,nrow);
    full_rhs=m_resize(full_rhs,nrow,(int) chars);
    fm_me=full_mat->me;
    fv_me=full_rhs->me;
    m_zero(full_mat);
    m_zero(full_rhs);
    np=tp->nodep;

/* calculate matrices */

    for(ip=0; ip<nrow; ip++)
	{
	p=np[ip+numsp1-1];
	do_q(p->next->back,ip,fm_me,fv_me);
	do_q(p->next->next->back,ip,fm_me,fv_me);
	}
    /*m_output(full_mat);*/
    /*m_output(full_rhs);*/

/* Cholesky factorization and solution for max likelihood est */
    ch_fact=m_copy(full_mat,MNULL);
    ch_fact=CHfactor(ch_fact);
    rhs=VNULL;
    for(ic=0; ic <chars; ic++)
	{
	rhs=get_col(full_rhs,ic,rhs);
    	mle=CHsolve(ch_fact,rhs,VNULL);
    	/*printf("\nMLE char %d:",ic);*/
    	/*v_output(mle);*/

/* set max_lik vals at nodes */
    	for(ip=0; ip<nrow; ip++)
	    {
	    p=np[ip+numsp1-1];
	    p->view[ic]=mle->ve[ip]; 
	    }
	}
    M_FREE(ch_fact);
    M_FREE(full_rhs);
    V_FREE(rhs);
    V_FREE(mle);
    }

void do_q(node *q,int ip, Real **fm_me,Real **fv_me)
    { /* add in values corresponding to one node */
    Real vinv;
    int ic,iq;
    vinv=2./ q->v;
    fm_me[ip][ip] += vinv;
    if(q->tip) 
	{
	for(ic=0; ic < chars; ic++)
	    fv_me[ip][ic] += q->view[ic]*vinv;
	}
    else {
	iq=(int) (q->number-numsp1);
	fm_me[ip][iq] -= vinv;
	fm_me[iq][ip] -= vinv;
	fm_me[iq][iq] += vinv;
	}
    }

