/* myops.c has code for some additional operations */
#include <stdio.h>
#include "matrix2.h"
#include "myops.h"
#include <math.h>

/* sv_add -- scalar-vector add  */
VEC	*sv_add(Real scalar,VEC *vector,VEC *out)
{
	int i,dim;
	Real *out_ve,*vec_ve;

	if ( vector==(VEC *)NULL )
		error(E_NULL,"sv_add");
	dim = vector->dim;
	if ( out==(VEC *)NULL || out->dim != dim )
		out = v_resize(out,dim);
	out_ve = out->ve;	vec_ve = vector->ve;
	for ( i=0; i<dim; i++ )
		out_ve[i] = scalar+vec_ve[i];
	return (out);
}

/* v_diff -- apply first difference to vector elements  */
VEC	*v_diff(VEC *vector,VEC *out)
{
	int i,dim;
	Real *out_ve,*vec_ve;

	if ( vector==(VEC *)NULL )
		error(E_NULL,"v_diff");
	if(vector->dim ==1) return((VEC *)NULL); /* no diff */
	dim = vector->dim -1;
	if ( out==(VEC *)NULL || out->dim != dim)
		out=v_resize(out,dim);
	out_ve = out->ve;	vec_ve = vector->ve;
	for ( i=0; i<dim; i++ )
		out_ve[i] = vec_ve[i+1]-vec_ve[i];
	return (out);
}

/* v_tr_r remove last vector element  */
VEC	*v_tr_r(VEC *vector,VEC *out)
{
	int i,dim;
	Real *out_ve,*vec_ve;

	if ( vector==(VEC *)NULL )
		error(E_NULL,"v_tr_r");
	if(vector->dim ==1) return((VEC *)NULL); /* no truncated vec */
	dim = vector->dim -1;
	if ( out==(VEC *)NULL || out->dim != dim)
		out=v_resize(out,dim);
	out_ve = out->ve;	vec_ve = vector->ve;
	for ( i=0; i<dim; i++ )
		out_ve[i] = vec_ve[i];
	return (out);
}

/* v_tr_r remove first vector element  */
VEC	*v_tr_l(VEC *vector,VEC *out)
{
	int i,dim;
	Real *out_ve,*vec_ve;

	if ( vector==(VEC *)NULL )
		error(E_NULL,"v_tr_l");
	if(vector->dim ==1) return((VEC *)NULL); /* no truncated vec */
	dim = vector->dim -1;
	if ( out==(VEC *)NULL || out->dim != dim)
		out=v_resize(out,dim);
	out_ve = out->ve;	vec_ve = vector->ve;
	for ( i=0; i<dim; i++ )
		out_ve[i] = vec_ve[i+1];
	return (out);
}

/* mnrand() -- generate normal random numbers */
Real mnrand(void)
{
	static int extra=0; /* switch to determine if a value is saved */
	static Real extra_val; /* saved value */
	static Real val1, val2,radsq,coeff; /* variables in circle */

	if(extra) 
		{
		extra=0;
		return(extra_val);
		}
	do
		{
		val1 = 2.*mrand() -1.;
		val2 = 2.*mrand() -1.;	
		radsq = val1*val1 + val2*val2;
		}
	while(radsq >=1.);

	coeff = sqrt(-2.*log(radsq)/radsq);
	extra_val=val1*coeff;
	extra=1;
	return(val2*coeff);	
}

/* v_nrand -- generate fake normal vector, mean 0, var sdev  */
VEC	*v_nrand(Real var, VEC *vector)
{
	int i,dim;
	Real sdev,*vec_ve;

	if ( vector==(VEC *)NULL )
		error(E_NULL,"v_nrand");
	sdev=sqrt(var);
	vec_ve=vector->ve;
	dim=vector->dim;
	for(i=0;i<dim;i++)
		vec_ve[i] = sdev* mnrand();
	return (vector);
}

/* v_lnorm -- generate fake log-normal vector  */
VEC	*v_lnorm(Real var, VEC *vector)
{
	int i,dim;
	Real sdev;

	if ( vector==(VEC *)NULL )
		error(E_NULL,"v_lnorm");
	sdev = sqrt(var);
	vector=v_nrand(sdev,vector); /* fake normal */
	vector = v_map(exp,vector,vector); /* exponentiate */
	return (vector);
}

/* double rec_sqrt(double x)
    {
	if( x >= 0.) return(1./sqrt(x));
	return(__infinity);
    }
*/

Real v_median(VEC *x, PERM *order)
	{ /* sort to obtain median, then restore x */
	Real med;
	int imed;
	if(order==PNULL) order=px_resize(order,x->dim);
	v_sort(x,order);
/*	printf("x sorted: order ");
	v_output(x); 
	px_output(order); 
*/
	imed = x->dim/2;
	if(x->dim != 2*imed) med=x->ve[imed];
	else med = .5*(x->ve[imed] + x->ve[imed-1]);
	x=pxinv_vec(order,x,x);
	return(med);
	}

/* v_adjoin -- combine two vectors; may be done in situ if vec1 == out */
VEC	*v_adjoin(vec1,vec2,out)
VEC	*vec1,*vec2,*out;
{
	u_int	dim;
	int v1dim,v2dim,i;
	Real *out_ve, *v2_ve, *v1_ve;

	if ( vec1==(VEC *)NULL || vec2==(VEC *)NULL )
		error(E_NULL,"v_add");
	dim = vec1->dim + vec2->dim;
	v1dim=vec1->dim; v2dim=vec2->dim;
	if ( out==(VEC *)NULL || out->dim != dim)
		out = v_resize(out,dim);
	out_ve=out->ve; v2_ve = vec2->ve; v1_ve=vec1->ve;
/*	out=_v_copy(vec1,out,0); */
	for(i=0;i<v1dim;i++) out_ve[i]=v1_ve[i];
	for(i=0;i<v2dim;i++) out_ve[v1dim+i] = v2_ve[i];
/*	out=_v_copy(vec2,out,v1dim); */
	return (out);
}

MAT *m_delete_row(MAT *in, int row,MAT *out)
    {
    Real **in_me;
    Real **out_me;
    int rnext, ir, ic;
    if(in==MNULL) error(E_NULL,"m_delete_row: null input matrix");
    if(in->m <= row || row <0)
	error(E_RANGE,"m_delete_row: bad row number");
    out = m_resize(out,in->m-1,in->n);
    in_me=in->me;
    out_me=out->me;
    rnext=0;
    for(ir=0; ir< in->m; ir++)
	if(ir != row)
	    {
	    for(ic=0; ic < in->n;ic++)
		out_me[rnext][ic]=in_me[ir][ic];
	    rnext++;
	    }
    return out;
    }

MAT *m_delete_col(MAT *in, int col,MAT *out)
    {
    Real **in_me;
    Real **out_me;
    int cnext, ic, ir;
    if(in==MNULL) error(E_NULL,"m_delete_col: null input matrix");
    if(in->n <= col || col <0)
	error(E_RANGE,"m_delete_col: bad col number");
    out = m_resize(out,in->m,in->n-1);
    in_me=in->me;
    out_me=out->me;
    cnext=0;
    for(ic=0; ic< in->n; ic++)
	if(ic != col)
	    {
	    for(ir=0; ir < in->m;ir++)
		out_me[ir][cnext]=in_me[ir][ic];
	    cnext++;
	    }
    return out;
    }

VEC *v_delete_elem(VEC *in, int row,VEC *out)
    {
    Real *in_ve;
    Real *out_ve;
    int rnext, ir;
    if(in==VNULL) error(E_NULL,"v_delete_elem: null input vector");
    if(in->dim <= row || row <0)
	error(E_RANGE,"v_delete_elem: bad row number");
    out=v_resize(out,in->dim-1);
    in_ve=in->ve;
    out_ve=out->ve;
    rnext=0;
    for(ir=0; ir< in->dim; ir++)
	if(ir != row)
	    {
	    out_ve[rnext]=in_ve[ir];
	    rnext++;
	    }
    return out;
    }

