#include "phylip.h"

/* Tool to print tree in file "treefile" to screen. Modified from RETREE
   version 3.52c ((c) Copyright 1993 by Joseph Felsenstein). */

#define maxsp           500   /* maximum number of species               */
#define maxsz           999   /* size of pointer array.  >= 2*maxsp - 1  */
                              /* this can be large without eating memory */
#define nmlngth         30    /* number of characters in species name    */

#define ibmpc0          false
#define ansi0           false
#define vt520           false
#define downn           2
#define overr           4

  typedef enum {    valid, remoov, quit    } reslttype;


typedef enum {
  horiz, vert, up, over, upcorner, downcorner, aa, cc, gg, tt, quest
  } chartype;
/* nodes will form a binary tree */

typedef struct node {       /* describes a tip species or an ancestor        */
  struct node *next, *back; /* pointers to nodes                             */
  long index;               /* number of the node                            */
  boolean tip;              /* present species are tips of tree              */
  boolean hasname;          /* true if tip has a name                        */
  char    name[nmlngth];    /* if it does, its name                          */
  boolean haslength;        /* true if node has a valid length               */
  double length;            /* length of this branch                         */
  double lbeyond;           /* distance beyond this node to most distant tip */
                            /* on left side                                  */
  double rbeyond;           /* distance beyond this node to most distant tip */
                            /* on right side                                 */
  long xcoord, ycoord;      /* used by printree                              */
  long ymin, ymax;
} node;

typedef node *pointarray[maxsz];
typedef enum {  arb, use, spec } howtree;
typedef enum {  left, down, upp, right } adjwindow;

typedef enum {
  rearr, transp, flipp, reroott, mpoint, lengthh, none
  } rearrtype;


typedef struct lenstore {
  long index;
  double length;
  boolean haslength;
} lenstore;

typedef struct midstoret {
  node *leftp, *rightp;
  double leftlength, rightlength;
} midstoret;


node *root, *garbage;
FILE *treefile, *outtree;
long spp, spp2, nonodes, outgrno, screenwidth, vscreenwidth,
     screenlines, col, treenumber, leftedge, topedge, treelines,
     hscroll, vscroll, scrollinc;
/* spp    = number of species
  nonodes = number of nodes in tree
  outgrno indicates outgroup */

double     backlength, backlength2, rlen1, llen1, rlen2, llen2, trweight;
boolean    waswritten, ibmpc, vt52, ansi, hasmult, haslengths,
           nolengths, nexus;
pointarray treenode;                /* pointers to all nodes in tree */

boolean    reversed[11];
boolean    graphic[11];
Char       cch[11];
howtree    how;
lenstore   oldls, templs;
midstoret  midstore;
char       intreename[100],outtreename[100];

long       atwhat, what, fromwhere, towhere, oldoutgrno;
boolean    subtree, written, oldwritten, restoring, wasleft, oldleft, readnext;
double     rtlength;
node      *nuroot;
rearrtype  lastop;
node      *washere;
Char      ch;


openfile(fp,filename,mode,application,perm)
     FILE **fp;
     char *filename;
     char *mode;
     char *application;
     char *perm;
{
  FILE *of;
  char file[100];
  strcpy(file,filename);
  while (1){
    of = fopen(file,mode);
    if (of)
      break;
    else {
      switch (*mode){
      case 'r':
        printf("%s:  can't read %s\n",application,file);
	file[0] = '\0';
        while (file[0] =='\0'){
          printf("Please enter a new filename>");
          gets(file);}
        break;
      case 'w':
        printf("%s: can't write %s\n",application,file);
	file[0] = '\0';
        while (file[0] =='\0'){
          printf("Please enter a new filename>");
          gets(file);}
        break;
      }
    }
  }
  *fp=of;
  if (perm != NULL)
    strcpy(perm,file);
}

void gnu(p)
     node **p;
{
  /* this and the following are do-it-yourself garbage collectors.
     Make a new node or pull one off the garbage list */
  if (garbage != NULL) {
    *p = garbage;
    garbage = garbage->next;
  } else
    *p = (node *)Malloc(sizeof(node));
  (*p)->next = NULL;
  (*p)->tip = false;
}  /* gnu */

void maketriad(p, index)
     node **p;
     long index;
{
  long i;
  node *q;
  q = NULL;
  for (i = 1; i <= 3; i++) {
    gnu(p);
    (*p)->index = index;
    (*p)->hasname = false;
    (*p)->haslength = false;
    (*p)->next = q;
    q = *p;
  }
  (*p)->next->next->next = *p;
  q = (*p)->next;
  while (*p != q) {
    (*p)->back = NULL;
    (*p)->tip = false;
    *p = (*p)->next;
  }
  treenode[index - 1] = *p;
}  /* maketriad */


void maketip(p, index)
     node **p;
     long index;
{
  gnu(p);
  (*p)->index = index;
  (*p)->tip = true;
  (*p)->hasname = false;
  (*p)->haslength = false;
  treenode[index - 1] = *p;
}  /* maketip */


void getoptions()
{
  /* interactively set options */
  Char ch;

  how = use;  /* user tree */
  outgrno = 1;

  if (scrollinc < screenwidth / 2.0)
    hscroll = scrollinc;
  else
    hscroll = screenwidth / 2;
  if (scrollinc < screenlines / 2.0)
    vscroll = scrollinc;
  else
    vscroll = screenlines / 2;
}  /* getoptions */

void configure()
{
  /* configure to machine -- set up special characters */
  chartype a;

  for (a = horiz; (long)a <= (long)quest; a = (chartype)((long)a + 1))
    reversed[(long)a] = false;
  for (a = horiz; (long)a <= (long)quest; a = (chartype)((long)a + 1))
    graphic[(long)a] = false;

  cch[(long)horiz] = '>';
  cch[(long)vert] = ' ';
  cch[(long)up] = '!';
  cch[(long)upcorner] = '+';
  cch[(long)downcorner] = '+';
  cch[(long)over] = '-';
}  /* configure */


void pregraph()
{
  /* turn on graphic characters */
  if (vt52)
    printf("\033F");
  if (ansi) {
    printf("\033(0");
    printf("\033[10m");
  }
}  /* pregraph */

void prereverse()
{
  /* turn on reverse video */
  if (vt52)
    printf("\033p");
  if (ansi)
    printf("\033[7m");
}  /* prereverse */


void prefix(a)
chartype a;
{
  /* give prefix appropriate for this character */
  if (reversed[(long)a])
    prereverse();
  if (graphic[(long)a])
    pregraph();
}  /* prefix */


void postgraph()
{
  /* turn off graphic characters */
  if (vt52)
    printf("\033G");
  if (ansi) {
    printf("\033[11m");
    printf("\033(B");
  }
}  /* postgraph */

void postreverse()
{
  /* turn off reverse video */
  if (vt52)
    printf("\033q");
  if (ansi)
    printf("\033[0m");
}  /* postreverse */


void postfix(a)
chartype a;
{
  /* give postfix appropriate for this character */
  if (reversed[(long)a])
    postreverse();
  if (graphic[(long)a])
    postgraph();
}  /* postfix */



void ltrav(p, localhl)
node *p;
boolean *localhl;
{
  node *q;

  if (p->tip) {
    (*localhl) = ((*localhl) && p->haslength);
    return;
  }
  q = p->next;
  do {
    (*localhl) = ((*localhl) && q->haslength);
    if ((*localhl))
      ltrav(q->back, localhl);
    q = q->next;
  } while (p != q);
}  /* ltrav */

boolean ifhaslengths()
{
  /* return true if every branch in tree has a length */
  /* Local variables for ifhaslengths: */
  boolean localhl;
  localhl = true;
  ltrav(root, &localhl);
  return localhl;
}  /* ifhaslengths */


void reroot(outgroup)
node *outgroup;
{
  /* reorients tree, putting outgroup in desired position. */
  node *p, *q, *newbottom, *oldbottom, *tempwashere;
  double outgrlength, temprtlength;

  if (!(outgroup->back->index != root->index || restoring))
    return;
  oldoutgrno = root->next->back->index;
  if (haslengths && !restoring)
    outgrlength = outgroup->length;
  if (restoring)
    outgroup = washere;
  newbottom = outgroup->back;
  p = treenode[newbottom->index - 1]->back;
  while (p->index != root->index) {
    oldbottom = treenode[p->index - 1];
    treenode[p->index - 1] = p;
    p = oldbottom->back;
  }
  p = root->next;
  q = root->next->next;
  tempwashere = root->next->next->back;
  if (haslengths)
    temprtlength = root->next->next->length;
  p->back->back = q->back;
  q->back->back = p->back;
  if (haslengths) {
    p->back->length = p->length + q->length;
    q->back->length = p->back->length;
  }
  if (restoring) {
    p->back = washere->back;
    p->back->back = p;
    q->back = washere;
    q->back->back = q;
    if (haslengths) {
      p->back->length -= rtlength;
      p->length = p->back->length;
      q->back->length = rtlength;
      q->length = rtlength;
      rtlength = temprtlength;
    }
  } else {
    p->back = outgroup;
    q->back = outgroup->back;
    outgroup->back->back = root->next->next;
    outgroup->back = root->next;
    if (haslengths) {
      root->next->length = outgrlength / 2.0;
      root->next->back->length = outgrlength / 2.0;
      root->next->next->length = outgrlength / 2.0;
      root->next->next->back->length = outgrlength / 2.0;
    }
  }
  washere = tempwashere;
  if (haslengths)
    rtlength = temprtlength;
  treenode[newbottom->index - 1] = newbottom;
}  /* reroot */


void ltrav_(p, lengthsum, lmin, tipmax, across,maxchar)
node *p;
double  lengthsum, lmin;
double *tipmax;
long   *across;
long   *maxchar;
{
  node *q;
  long rchar, nl;
  double sublength;

  if (p->tip) {
    if (lengthsum > (*tipmax))
      (*tipmax) = lengthsum;
    if (lmin == 0.0)
      return;
    rchar = (long)(lengthsum / (*tipmax) * (*across) + 0.5);
    nl = strlen(treenode[p->index - 1]->name);
    if (rchar + nl > (*maxchar))
      (*across) = (*maxchar) - (long)(nl * (*tipmax) / lengthsum + 0.5);
    return;
  }
  q = p->next;
  do {
    if (q->length >= lmin)
      sublength = q->length;
    else
      sublength = lmin;
    ltrav_(q->back, lengthsum + sublength, lmin,tipmax,across,maxchar);
    q = q->next;
  } while (p != q);
}


void precoord(nuroot, subtree,tipmax,across)
node    *nuroot;
boolean *subtree;
double  *tipmax;
long    *across;
{
  /*                                                              *
   *  set tipmax and across so that tree is scaled to screenwidth */

  double oldtipmax, minimum;
  long i;
  long maxchar;

  (*tipmax) = 0.0;
  if ((*subtree))
    maxchar = vscreenwidth - 13;
  else
    maxchar = vscreenwidth - 5;
  (*across) = maxchar;
  ltrav_(nuroot, 0.0, 0.0, tipmax,across,&maxchar);
  i = 0;
  do {
    oldtipmax = (*tipmax);
    minimum = 3.0 / (*across) * (*tipmax);
    ltrav_(nuroot, 0.0, minimum,tipmax,across,&maxchar);
    i++;
  } while (fabs((*tipmax) - oldtipmax) > 0.01 * oldtipmax && i <= 40);
}  /* precoord */

void coordinates(p, lengthsum, across, tipy, tipmax)
node   *p;
double lengthsum;
long   *across,*tipy;
double *tipmax;
{
  /* establishes coordinates of nodes for display with lengths */
  node *q, *first, *last;

  if (p->tip) {
    p->xcoord = (long)((*across) * lengthsum / (*tipmax) + 0.5);
    p->ycoord = (*tipy);
    p->ymin   = (*tipy);
    p->ymax   = (*tipy);
    (*tipy)  += downn;
    return;
  }
  q = p->next;
  do {
   coordinates(q->back, lengthsum + q->length, across, tipy, tipmax);
    q = q->next;
  } while (p != q);
  first = p->next->back;
  q = p;
  while (q->next != p)
    q = q->next;
  last = q->back;
  p->xcoord = (long)((*across) * lengthsum / (*tipmax) + 0.5);
  p->ycoord = (first->ycoord + last->ycoord) / 2;
  p->ymin = first->ymin;
  p->ymax = last->ymax;
}  /* coordinates */

void flatcoordinates(p, tipy)
node *p;
long *tipy;
{
  /* establishes coordinates of nodes for display without lengths */
  node *q, *first, *last;

  if (p->tip) {
    p->xcoord = 0;
    p->ycoord = (*tipy);
    p->ymin   = (*tipy);
    p->ymax   = (*tipy);
    (*tipy) += downn;
    return;
  }
  q = p->next;
  do {
    flatcoordinates(q->back, tipy);
    q = q->next;
  } while (p != q);
  first = p->next->back;
  q = p->next;
  while (q->next != p)
    q = q->next;
  last = q->back;
  p->xcoord = (last->ymax - first->ymin) * 3 / 2;
  p->ycoord = (first->ycoord + last->ycoord) / 2;
  p->ymin = first->ymin;
  p->ymax = last->ymax;
}  /* flatcoordinates */


void grwrite(c, num, pos)
chartype c;
long num;
long *pos;
{
  long i;

  prefix(c);
  for (i = 1; i <= num; i++) {
    if ((*pos) >= leftedge && (*pos) - leftedge + 1 < screenwidth)
      putchar(cch[(long)c]);
    (*pos)++;
  }
  postfix(c);
}

void chwrite(ch, num, pos)
Char ch;
long num;
long *pos;
{
  long i;

  for (i = 1; i <= num; i++) {
    if ((*pos) >= leftedge && (*pos) - leftedge + 1 < screenwidth)
      putchar(ch);
    (*pos)++;
  }
}

void nnwrite(nodenum, num, pos)
long nodenum, num;
long *pos;
{
  int i;
  long leftx;

  leftx = leftedge - (*pos);
  if ((*pos) >= leftedge && (*pos) - leftedge + num < screenwidth)
    printf("%*ld", (int)num, nodenum);
  else if (leftx > 0 && leftx < 3)
    for(i=0;i<(int)(num-leftx);i++)
      printf(" ");
  (*pos) += num;
}

void stwrite(s, length, pos)
Char *s;
long length;
long *pos;
{
  if ((*pos) >= leftedge && (*pos) - leftedge + 1 < screenwidth)
    printf("%*s", (int)length, s);
  (*pos) += length;
}

void drawline(i, nuroot, subtree)
long i;
node *nuroot;
boolean *subtree;
{
  /* draws one row of the tree diagram by moving up tree */
  long pos;
  node *p, *q, *r, *first, *last;
  long n, j;
  boolean extra, done;
  chartype c, d;
  pos = 1;
  p = nuroot;
  q = nuroot;

  extra = false;
  if (i == p->ycoord && (p == root || (*subtree))) {
    c = over;
    if ((*subtree))
      stwrite("Subtree:", 8L, &pos);
    if (p->index >= 100)
      nnwrite(p->index, 3L, &pos);
    else if (p->index >= 10) {
      grwrite(c, 1L, &pos);
      nnwrite(p->index, 2L, &pos);
    } else {
      grwrite(c, 2L, &pos);
      nnwrite(p->index, 1L, &pos);
    }
    extra = true;
  } else {
    if ((*subtree))
      stwrite("          ", 10L, &pos);
    else
      stwrite("  ", 2L, &pos);
  }
  do {
    if (!p->tip) {
      r = p->next;
      done = false;
      do {
	if (i >= r->back->ymin && i <= r->back->ymax) {
	  q = r->back;
	  done = true;
	}
	r = r->next;
      } while (!(done || r == p));
      first = p->next->back;
      r = p->next;
      while (r->next != p)
	r = r->next;
      last = r->back;
    }
    done = (p == q);
    if (haslengths && !nolengths)
      n = q->xcoord - p->xcoord;
    else
      n = p->xcoord - q->xcoord;
    if (n < 3 && !q->tip)
      n = 3;
    if (extra) {
      n--;
      extra = false;
    }
    if (q->ycoord == i && !done) {
      if (q->ycoord > p->ycoord)
	d = upcorner;
      else
	d = downcorner;
      c = over;
      if (!haslengths && !q->haslength)
	c = horiz;
      if (n > 1 || q->tip) {
	grwrite(d, 1L, &pos);
	grwrite(c, n - 3, &pos);
      }
      if (q->index >= 100)
	nnwrite(q->index, 3L, &pos);
      else if (q->index >= 10) {
	grwrite(c, 1L, &pos);
	nnwrite(q->index, 2L, &pos);
      } else {
	grwrite(c, 2L, &pos);
	nnwrite(q->index, 1L, &pos);
      }
      extra = true;
    } else if (!q->tip) {
      if (last->ycoord > i && first->ycoord < i && i != p->ycoord) {
	c = up;
	grwrite(c, 1L, &pos);
	chwrite(' ', n - 1, &pos);
      } else
	chwrite(' ', n, &pos);
    } else
      chwrite(' ', n, &pos);
    if (p != q)
      p = q;
  } while (!done);
  if (p->ycoord == i && p->tip) {
    if (p->hasname) {
      n = 0;
      for (j = 1; j <= nmlngth; j++) {
	if (treenode[p->index - 1]->name[j - 1] != '\0')
	  n = j;
      }
      chwrite(':', 1L, &pos);
      for (j = 0; j < n; j++)
	chwrite(treenode[p->index - 1]->name[j], 1L, &pos);
    }
  }
  putchar('\n');
}  /* drawline */

void printree()
{
  /* prints out diagram of the tree */
  /* Local variables for printree:  */
  long across, tipy;
  double tipmax;
  long i, rover, dow, vmargin;

  haslengths = ifhaslengths();
  if (!subtree)
    nuroot = root;
  printf((ansi || ibmpc) ? "\033[2J\033[H" :
         (vt52)          ? "\033E\033H"    : "\n");
  tipy = 1;
  rover = 100 / spp;
  if (rover > overr)
    rover = overr;
  dow = downn;
  if (spp * dow > screenlines && !subtree) {
    dow--;
    rover--;
  }
  if (haslengths && !nolengths) {
    precoord(nuroot, &subtree,&tipmax,&across);
    coordinates(nuroot, 0.0, &across,&tipy,&tipmax);
  } else
    flatcoordinates(nuroot, &tipy);

  vmargin = 2;
  treelines = tipy - dow;
  if (topedge != 1) {
    printf("** %ld lines above screen **\n", topedge - 1);
    vmargin++;
  }
  if ((treelines - topedge + 1) > (screenlines - vmargin))
    vmargin++;
  for (i = 1; i <= treelines; i++) {
    if (i >= topedge && i < topedge + screenlines - vmargin)
      drawline(i, nuroot,&subtree);
  }
  if ((treelines - topedge + 1) > (screenlines - vmargin)) {
    printf("** %ld", treelines - (topedge - 1 + screenlines - vmargin));
    printf(" lines below screen **\n");
  }
  if (treelines - topedge + vmargin + 1 < screenlines)
    putchar('\n');
}  /* printree */

#undef downn
#undef overr

void togglelengths()
{
  nolengths = !nolengths;
  printree();
}

void getch(c)
Char *c;
{
  /* get next nonblank character */
  do {
    if (eoln(treefile)) {
      fscanf(treefile, "%*[^\n]");
      getc(treefile);
    }
    *c = getc(treefile);
    if (*c == '\n')
      *c = ' ';
  } while (*c == ' ');
}  /* getch */

void findch(c)
Char c;
{
  /* scan forward until find character c */
  boolean done;

  done = false;
  while (!(done)) {
    if (c == ',') {
      if (ch == '(' || ch == ')' || ch == ';') {
	printf("\nERROR IN USER TREE: UNMATCHED PARENTHESIS OR MISSING COMMA\n");
	exit(-1);
      } else if (ch == ',')
	done = true;
    } else if (c == ')') {
      if (ch == '(' || ch == ',' || ch == ';') {
	printf("\nERROR IN USER TREE: UNMATCHED PARENTHESIS OR NOT BIFURCATED NODE\n");
	exit(-1);
      } else {
	if (ch == ')')
	  done = true;
      }
    } else if (c == ';') {
      if (ch != ';') {
	printf("\nERROR IN USER TREE: UNMATCHED PARENTHESIS OR MISSING SEMICOLON\n");
	exit(-1);
      } else
	done = true;
    }
    if (ch != ')')
      continue;
    if (eoln(treefile)) {
      fscanf(treefile, "%*[^\n]");
      getc(treefile);
    }
    ch = getc(treefile);
    if (ch == '\n')
      ch = ' ';
  }
}  /* findch */


void processlength(p)
node *p;
{
int i=0;
char tmp[100];

do {(getch(&tmp[i]));}
  while (strchr("0123456789-.+-",tmp[i++]) != NULL);
ch = tmp[i-1];
tmp[--i]=0;
#ifdef VMS
sscanf(tmp,"%lf",&(p->length));
#else
p->length    = atof(tmp);
#endif
p->haslength = true;
}


void addelement(p, endoffile,nextnode,lparens,n)
node **p;
boolean *endoffile;
long *nextnode,*lparens,*n;
{
  /* recursive procedure adds nodes to user-defined tree */
  node *q, *x, *temp;
  long i;
  char name[100];

  do {
    (*endoffile) = eof(treefile);
    if (!(*endoffile)) {
      if (eoln(treefile)) {
	fscanf(treefile, "%*[^\n]");
	getc(treefile);
      }
    }
    (*endoffile) = eof(treefile);
    if (!(*endoffile)) {
      ch = getc(treefile);
      if (ch == '\n')
	ch = ' ';
    }
  } while (!(ch != ' ' || (*endoffile)));
  if (*endoffile)
    exit(-1);
  if (*p == root && (*endoffile))
    printf("ERROR IN USER TREE: INPUT FILE TRUNCATED\n");
  if (ch == '(' ) {
    if ((*lparens) >= spp - 1) {
      printf("\nERROR IN USER TREE: TOO MANY LEFT PARENTHESES\n");
      exit(-1);
    } else {
      (*nextnode)++;
      (*lparens)++;
      maketriad(&q, (*nextnode));
      addelement(&q->next->back,endoffile,nextnode,lparens,n);
      q->next->back->back = q->next;
      q->next->haslength = q->next->back->haslength;
      if (q->next->haslength)
	q->next->length = q->next->back->length;
      findch(',');
      addelement(&q->next->next->back,endoffile,nextnode,lparens,n);
      q->next->next->back->back = q->next->next;
      q->next->next->haslength = q->next->next->back->haslength;
      if (q->next->next->haslength)
	q->next->next->length = q->next->next->back->length;
      if (ch == ',') {
	if (*p == root) {
	  (*nextnode)++;
	  maketriad(&x, (*nextnode));
	  temp = q->next->next->back;
	  x->back = q->next->next;
	  q->next->next->back = x;
	  temp->back = x->next;
	  x->next->back = temp;
	  x->next->length = x->next->back->length;
	  x->next->haslength = x->next->back->haslength;
	  x->haslength = true;
	  x->back->haslength = true;
	  x->length = 0.0;
	  x->back->length = 0.0;
	  addelement(&x->next->next->back,endoffile,nextnode,lparens,n);
	  x->next->next->back->back = x->next->next;
	  x->next->next->haslength = x->next->next->back->haslength;
	  if (x->next->next->haslength)
	    x->next->next->length = x->next->next->back->length;
	} else {
	  printf("ERROR IN USER TREE: TRIFURCATION NOT ALLOWED ");
	  printf("EXCEPT AT ROOT\n");
	  exit(-1);
	}
      }
      findch(')');
      *p = q;
    }
  } else {
    spp2++;
     for (i=0;i<nmlngth;++i)
       name[i] = ' ';
    (*n) = 1;
    while (ch != ',' && ch != ')' && ch != ':' && (*n) <= nmlngth) {
      if (ch == '_')
	ch = ' ';
      name[(*n) - 1] = ch;
      name[(*n)] = '\0';
      if (eoln(treefile)) {
	fscanf(treefile, "%*[^\n]");
	getc(treefile);
      }
      ch = getc(treefile);
      if (ch == '\n')
	ch = ' ';
      (*n)++;
    }
    maketip(p, spp2);
    if ((*n) > 1){
      (*p)->hasname = true;
      strncpy(treenode[spp2 - 1]->name,name,nmlngth);
    }
  }
  if (ch == ':')
    processlength(*p);
}  /* addelement */


void treeread()
{
  /* read in user-defined tree and set it up */
  long n, nextnode, lparens;
  boolean endoffile;
  long i;

  hasmult = false;
  if (!readnext)
    openfile(&treefile,TREEFILE,"r","retree",intreename);
  spp2 = 0;
  spp = maxsp;
  nonodes = spp * 2 - 1;
  maketriad(&root, spp + 1);
  nextnode = spp;
  root->back = NULL;
  lparens = 0;
  addelement(&root, &endoffile,&nextnode,&lparens,&n);
  if (ch == '[' ) {
    if (!eoln(treefile)) {
      hasmult = true;
      fscanf(treefile, "%lf", &trweight);
     getch(&ch);
      if (ch != ']') {
	printf("ERROR: MISSING RIGHT SQUARE BRACKET\n");
	exit(-1);
      } else {
	getch(&ch);
	if (ch != ';') {
	  printf("ERROR: MISSING SEMICOLON AFTER SQUARE BRACKETS\n");
	  exit(-1);
	}
      }
    }
  }
  findch(';');
  for (i = 1; i < spp2; i++) {
    treenode[spp2 + i - 1] = treenode[spp + i - 1];
    treenode[spp2 + i - 1]->index = spp2 + i;
    treenode[spp2 + i - 1]->next->index = spp2 + i;
    treenode[spp2 + i - 1]->next->next->index = spp2 + i;
  }
  spp = spp2;
  nonodes = spp * 2 - 1;
  root = treenode[spp];
  printf("\n\n");
}  /* treeread */

void buildtree()
{
  long i;
  printf("\nReading tree file ...\n\n");
  treeread();
  outgrno = root->next->back->index;
  reroot(treenode[outgrno - 1]);
}  /* buildtree */

void treeconstruct()
{
  /* constructs a binary tree from the pointers in treenode. */

  waswritten = false;
  readnext = false;
  restoring = false;
  subtree = false;
  topedge = 1;
  leftedge = 1;
  buildtree();
  readnext = false;
  written = false;
  printree();
  lastop = none;
}  /* treeconstruct */


main(argc, argv)
     int argc;
     Char *argv[];
{
  /* reads in spp. Then calls treeconstruct to                               *
     construct the tree and query the user
                             */

#ifdef MAC
  macsetup("Retree","");
#endif
  nexus     = false;
  nolengths = false;
  scrollinc = 20;
  screenlines = 1000;
  screenwidth = 1000;
  vscreenwidth = 80;
  ibmpc = ibmpc0;
  ansi  = ansi0;
  vt52  = vt520;
  getoptions();
  configure();
  treeconstruct();
  if (waswritten)
    FClose(outtree);
  FClose(treefile);
  FClose(outtree);
#ifdef MAC
  fixmacfile(intreename);
  fixmacfile(outtreename);
#endif
  exit(0);
}  /* Retree */

int eof(f)
     FILE *f;
{
  register int ch;

  if (feof(f))
    return 1;
  if (f == stdin)
    return 0;
    ch = getc(f);
  if (ch == EOF)
    return 1;
  ungetc(ch, f);
  return 0;
}


int eoln(f)
     FILE *f;
{
  register int ch;

  ch = getc(f);
    if (ch == EOF)
      return 1;
  ungetc(ch, f);
  return (ch == '\n');
}

void memerror()
{
  printf("Error allocating memory\n");
  exit(-1);
}

MALLOCRETURN *mymalloc(x)
     long  x;
{
  MALLOCRETURN *mem;
  mem = (MALLOCRETURN *)calloc(1,x);
  if (!mem)
    memerror();
  else
    return (MALLOCRETURN *)mem;
}


